@@ -1,10 +1,10 @@
 import fnmatch
 import tempfile
-
 from contextlib import contextmanager
-from json import load
-from os import makedirs
-from os import unlink
+from os import (
+    makedirs,
+    unlink,
+)
 from os.path import (
     abspath,
     basename,
@@ -13,23 +13,34 @@ from os.path import (
     join,
     sep,
 )
-from re import compile, escape
+from re import (
+    compile,
+    escape,
+)
+from typing import (
+    Any,
+    Dict,
+    List,
+    Type,
+)
+from urllib.parse import urlencode
 
 from galaxy.util.bunch import Bunch
 
 from .config_util import read_file
-from .transport import get_file
-from .transport import post_file
 from .transport import (
+    get_file,
+    post_file,
     rsync_get_file,
     rsync_post_file,
     scp_get_file,
     scp_post_file,
 )
-from .util import copy_to_path
-from .util import directory_files
-from .util import unique_path_prefix
-
+from .util import (
+    copy_to_path,
+    directory_files,
+    unique_path_prefix,
+)
 
 DEFAULT_MAPPED_ACTION = 'transfer'  # Not really clear to me what this should be, exception?
 DEFAULT_PATH_MAPPER_TYPE = 'prefix'
@@ -47,9 +58,11 @@ path_type = Bunch(
     CONFIG="config",
     # Files from tool's tool_dir (for now just wrapper if available).
     TOOL="tool",
-    # Input work dir files - e.g.task-split input file
+    # Input tool work dir files - e.g. task-split input file
     WORKDIR="workdir",
-    # Input work dir files - e.g. metadata files, etc..
+    # Job directory files (e.g. tool standard input/output and containerized command).
+    JOBDIR="jobdir",
+    # Input metadata dir files - e.g. metadata files, etc..
     METADATA="metadata",
     # Galaxy output datasets in their final home.
     OUTPUT="output",
@@ -58,8 +71,10 @@ path_type = Bunch(
     # Meta job and data files (e.g. Galaxy metadata generation files and
     # metric instrumentation files)
     OUTPUT_METADATA="output_metadata",
+    # Job directory files output.
+    OUTPUT_JOBDIR="output_jobdir",
     # Other fixed tool parameter paths (likely coming from tool data, but not
-    # nessecarily). Not sure this is the best name...
+    # necessarily).
     UNSTRUCTURED="unstructured",
 )
 
@@ -69,10 +84,12 @@ ACTION_DEFAULT_PATH_TYPES = [
     path_type.CONFIG,
     path_type.TOOL,
     path_type.WORKDIR,
+    path_type.JOBDIR,
     path_type.METADATA,
     path_type.OUTPUT,
     path_type.OUTPUT_WORKDIR,
     path_type.OUTPUT_METADATA,
+    path_type.OUTPUT_JOBDIR,
 ]
 ALL_PATH_TYPES = ACTION_DEFAULT_PATH_TYPES + [path_type.UNSTRUCTURED]
 
@@ -80,7 +97,7 @@ MISSING_FILES_ENDPOINT_ERROR = "Attempted to use remote_transfer action without
 MISSING_SSH_KEY_ERROR = "Attempt to use file transfer action requiring an SSH key without specifying a ssh_key."
 
 
-class FileActionMapper(object):
+class FileActionMapper:
     """
     Objects of this class define how paths are mapped to actions.
 
@@ -99,53 +116,62 @@ class FileActionMapper(object):
     ...     mock_client = Bunch(default_file_action=default_action, action_config_path=f.name, files_endpoint=None)
     ...     mapper = FileActionMapper(mock_client)
     ...     as_dict = config=mapper.to_dict()
-    ...     # print(as_dict["paths"])
     ...     mapper = FileActionMapper(config=as_dict) # Serialize and deserialize it to make sure still works
     ...     unlink(f.name)
     ...     return mapper
     >>> mapper = mapper_for(default_action='none', config_contents=json_string)
     >>> # Test first config line above, implicit path prefix mapper
-    >>> action = mapper.action('/opt/galaxy/tools/filters/catWrapper.py', 'input')
+    >>> action = mapper.action({'path': '/opt/galaxy/tools/filters/catWrapper.py'}, 'input')
     >>> action.action_type == u'none'
     True
     >>> action.staging_needed
     False
     >>> # Test another (2nd) mapper, this one with a different action
-    >>> action = mapper.action('/galaxy/data/files/000/dataset_1.dat', 'input')
+    >>> action = mapper.action({'path': '/galaxy/data/files/000/dataset_1.dat'}, 'input')
     >>> action.action_type == u'transfer'
     True
     >>> action.staging_needed
     True
     >>> # Always at least copy work_dir outputs.
-    >>> action = mapper.action('/opt/galaxy/database/working_directory/45.sh', 'workdir')
+    >>> action = mapper.action({'path': '/opt/galaxy/database/working_directory/45.sh'}, 'workdir')
     >>> action.action_type == u'copy'
     True
     >>> action.staging_needed
     True
     >>> # Test glob mapper (matching test)
-    >>> mapper.action('/cool/bamfiles/projectABC/study1/patient3.bam', 'input').action_type == u'copy'
+    >>> mapper.action({'path': '/cool/bamfiles/projectABC/study1/patient3.bam'}, 'input').action_type == u'copy'
     True
     >>> # Test glob mapper (non-matching test)
-    >>> mapper.action('/cool/bamfiles/projectABC/study1/patient3.bam.bai', 'input').action_type == u'none'
+    >>> mapper.action({'path': '/cool/bamfiles/projectABC/study1/patient3.bam.bai'}, 'input').action_type == u'none'
     True
     >>> # Regex mapper test.
-    >>> mapper.action('/old/galaxy/data/dataset_10245.dat', 'input').action_type == u'copy'
+    >>> mapper.action({'path': '/old/galaxy/data/dataset_10245.dat'}, 'input').action_type == u'copy'
     True
     >>> # Doesn't map unstructured paths by default
-    >>> mapper.action('/old/galaxy/data/dataset_10245.dat', 'unstructured').action_type == u'none'
+    >>> mapper.action({'path': '/old/galaxy/data/dataset_10245.dat'}, 'unstructured').action_type == u'none'
     True
     >>> input_only_mapper = mapper_for(default_action="none", config_contents=r'''{"paths": [ \
       {"path": "/", "action": "transfer", "path_types": "input"} \
     ] }''')
-    >>> input_only_mapper.action('/dataset_1.dat', 'input').action_type == u'transfer'
+    >>> input_only_mapper.action({'path': '/dataset_1.dat'}, 'input').action_type == u'transfer'
     True
-    >>> input_only_mapper.action('/dataset_1.dat', 'output').action_type == u'none'
+    >>> input_only_mapper.action({'path': '/dataset_1.dat'}, 'output').action_type == u'none'
     True
     >>> unstructured_mapper = mapper_for(default_action="none", config_contents=r'''{"paths": [ \
       {"path": "/", "action": "transfer", "path_types": "*any*"} \
     ] }''')
-    >>> unstructured_mapper.action('/old/galaxy/data/dataset_10245.dat', 'unstructured').action_type == u'transfer'
+    >>> unstructured_mapper.action({'path': '/old/galaxy/data/dataset_10245.dat'}, 'unstructured').action_type == u'transfer'
     True
+    >>> match_type_only_mapper = mapper_for(default_action="none", config_contents=r'''{"paths": [ \
+      {"action": "transfer", "path_types": "input"}, \
+      {"action": "remote_copy", "path_types": "output"} \
+    ] }''')
+    >>> input_action = match_type_only_mapper.action({}, 'input')
+    >>> input_action.action_type
+    'transfer'
+    >>> output_action = match_type_only_mapper.action({}, 'output')
+    >>> output_action.action_type
+    'remote_copy'
     """
 
     def __init__(self, client=None, config=None):
@@ -162,7 +188,8 @@ class FileActionMapper(object):
         self.mappers = mappers_from_dicts(config.get("paths", []))
         self.files_endpoint = config.get("files_endpoint", None)
 
-    def action(self, path, type, mapper=None):
+    def action(self, source, type, mapper=None):
+        path = source.get("path", None)
         mapper = self.__find_mapper(path, type, mapper)
         action_class = self.__action_class(path, type, mapper)
         file_lister = DEFAULT_FILE_LISTER
@@ -170,7 +197,7 @@ class FileActionMapper(object):
         if mapper:
             file_lister = mapper.file_lister
             action_kwds = mapper.action_kwds
-        action = action_class(path, file_lister=file_lister, **action_kwds)
+        action = action_class(source, file_lister=file_lister, **action_kwds)
         self.__process_action(action, type)
         return action
 
@@ -196,7 +223,7 @@ class FileActionMapper(object):
         if action_config_path:
             config = read_file(action_config_path)
         else:
-            config = dict()
+            config = getattr(client, "file_actions", {})
         config["default_action"] = client.default_file_action
         config["files_endpoint"] = client.files_endpoint
         for attr in ['ssh_key', 'ssh_user', 'ssh_port', 'ssh_host']:
@@ -204,13 +231,12 @@ class FileActionMapper(object):
                 config[attr] = getattr(client, attr)
         return config
 
-    def __load_action_config(self, path):
-        config = load(open(path, 'rb'))
-        self.mappers = mappers_from_dicts(config.get('paths', []))
-
     def __find_mapper(self, path, type, mapper=None):
         if not mapper:
-            normalized_path = abspath(path)
+            if path is not None:
+                normalized_path = abspath(path)
+            else:
+                normalized_path = None
             for query_mapper in self.mappers:
                 if query_mapper.matches(normalized_path, type):
                     mapper = query_mapper
@@ -221,8 +247,8 @@ class FileActionMapper(object):
         action_type = self.default_action if type in ACTION_DEFAULT_PATH_TYPES else "none"
         if mapper:
             action_type = mapper.action_type
-        if type in ["workdir", "output_workdir", "output_metadata"] and action_type == "none":
-            # We are changing the working_directory relative to what
+        if type in ["workdir", "jobdir", "output_workdir", "output_metadata", "output_jobdir"] and action_type == "none":
+            # We are changing the working_directory/job_directory relative to what
             # Galaxy would use, these need to be copied over.
             action_type = "copy"
         action_class = actions.get(action_type, None)
@@ -247,9 +273,8 @@ class FileActionMapper(object):
             raise Exception(MISSING_FILES_ENDPOINT_ERROR)
         if "?" not in url_base:
             url_base = "%s?" % url_base
-        # TODO: URL encode path.
-        url = "%s&path=%s&file_type=%s" % (url_base, action.path, file_type)
-        action.url = url
+        url_params = urlencode({"path": action.path, "file_type": file_type})
+        action.url = f"{url_base}{url_params}"
 
     def __inject_ssh_properties(self, action):
         for attr in ["ssh_key", "ssh_host", "ssh_port", "ssh_user"]:
@@ -266,13 +291,19 @@ REQUIRED_ACTION_KWD = object()
 UNSET_ACTION_KWD = "__UNSET__"
 
 
-class BaseAction(object):
-    action_spec = {}
+class BaseAction:
+    whole_directory_transfer_supported = False
+    action_spec: Dict[str, Any] = {}
+    action_type: str
 
-    def __init__(self, path, file_lister=None):
-        self.path = path
+    def __init__(self, source, file_lister=None):
+        self.source = source
         self.file_lister = file_lister or DEFAULT_FILE_LISTER
 
+    @property
+    def path(self):
+        return self.source.get("path")
+
     def unstructured_map(self, path_helper):
         unstructured_map = self.file_lister.unstructured_map(self.path)
         if self.staging_needed:
@@ -297,14 +328,30 @@ class BaseAction(object):
     def staging_action_local(self):
         return self.staging == STAGING_ACTION_LOCAL
 
+    def _extend_base_dict(self, **kwds):
+        base_dict = dict(
+            path=self.path,  # For older Pulsar servers (pre-0.13.0?)
+            source=self.source,
+            action_type=self.action_type,
+        )
+        base_dict.update(**kwds)
+        return base_dict
+
     def to_dict(self):
-        return dict(action_type=self.action_type)
+        return self._extend_base_dict()
 
     def __str__(self):
         as_dict = self.to_dict()
         attribute_str = ""
+        first = True
         for key, value in as_dict.items():
-            attribute_str += "%s=%s" % (key, value)
+            if key == "source":
+                continue
+            if first:
+                first = False
+            else:
+                attribute_str += ","
+            attribute_str += "{}={}".format(key, value)
         return "FileAction[%s]" % attribute_str
 
 
@@ -317,11 +364,11 @@ class NoneAction(BaseAction):
     staging = STAGING_ACTION_NONE
 
     def to_dict(self):
-        return dict(path=self.path, action_type=self.action_type)
+        return self._extend_base_dict()
 
     @classmethod
     def from_dict(cls, action_dict):
-        return NoneAction(path=action_dict["path"])
+        return NoneAction(source=action_dict["source"])
 
     def path_rewrite(self, path_helper, path=None):
         return None
@@ -338,16 +385,13 @@ class RewriteAction(BaseAction):
     action_type = "rewrite"
     staging = STAGING_ACTION_NONE
 
-    def __init__(self, path, file_lister=None, source_directory=None, destination_directory=None):
-        self.path = path
-        self.file_lister = file_lister or DEFAULT_FILE_LISTER
+    def __init__(self, source, file_lister=None, source_directory=None, destination_directory=None):
+        super().__init__(source, file_lister=file_lister)
         self.source_directory = source_directory
         self.destination_directory = destination_directory
 
     def to_dict(self):
-        return dict(
-            path=self.path,
-            action_type=self.action_type,
+        return self._extend_base_dict(
             source_directory=self.source_directory,
             destination_directory=self.destination_directory,
         )
@@ -355,7 +399,7 @@ class RewriteAction(BaseAction):
     @classmethod
     def from_dict(cls, action_dict):
         return RewriteAction(
-            path=action_dict["path"],
+            source=action_dict["source"],
             source_directory=action_dict["source_directory"],
             destination_directory=action_dict["destination_directory"],
         )
@@ -392,12 +436,9 @@ class RemoteCopyAction(BaseAction):
     action_type = "remote_copy"
     staging = STAGING_ACTION_REMOTE
 
-    def to_dict(self):
-        return dict(path=self.path, action_type=self.action_type)
-
     @classmethod
     def from_dict(cls, action_dict):
-        return RemoteCopyAction(path=action_dict["path"])
+        return RemoteCopyAction(source=action_dict["source"])
 
     def write_to_path(self, path):
         copy_to_path(open(self.path, "rb"), path)
@@ -421,16 +462,16 @@ class RemoteTransferAction(BaseAction):
     action_type = "remote_transfer"
     staging = STAGING_ACTION_REMOTE
 
-    def __init__(self, path, file_lister=None, url=None):
-        super(RemoteTransferAction, self).__init__(path, file_lister=file_lister)
+    def __init__(self, source, file_lister=None, url=None):
+        super().__init__(source, file_lister=file_lister)
         self.url = url
 
     def to_dict(self):
-        return dict(path=self.path, action_type=self.action_type, url=self.url)
+        return self._extend_base_dict(url=self.url)
 
     @classmethod
     def from_dict(cls, action_dict):
-        return RemoteTransferAction(path=action_dict["path"], url=action_dict["url"])
+        return RemoteTransferAction(source=action_dict["source"], url=action_dict["url"])
 
     def write_to_path(self, path):
         get_file(self.url, path)
@@ -439,6 +480,33 @@ class RemoteTransferAction(BaseAction):
         post_file(self.url, pulsar_path)
 
 
+class RemoteObjectStoreCopyAction(BaseAction):
+    """
+    """
+    action_type = "remote_object_store_copy"
+    staging = STAGING_ACTION_REMOTE
+    inject_object_store = True
+
+    @classmethod
+    def from_dict(cls, action_dict):
+        return RemoteObjectStoreCopyAction(source=action_dict["source"])
+
+    def write_to_path(self, path):
+        assert self.object_store  # Make sure object_store attribute injected
+        assert "object_store_ref" in self.source
+        object_store_ref = self.source["object_store_ref"]
+        dataset_object = Bunch(
+            id=object_store_ref["dataset_id"],
+            uuid=object_store_ref["dataset_uuid"],
+            object_store_id=object_store_ref["object_store_id"],
+        )
+        filename = self.object_store.get_filename(dataset_object)
+        copy_to_path(open(filename, 'rb'), path)
+
+    def write_from_path(self, pulsar_path):
+        raise NotImplementedError("Writing raw files to object store not supported at this time.")
+
+
 class PubkeyAuthenticatedTransferAction(BaseAction):
     """Base class for file transfers requiring an SSH public/private key
     """
@@ -451,18 +519,20 @@ class PubkeyAuthenticatedTransferAction(BaseAction):
     )
     staging = STAGING_ACTION_REMOTE
 
-    def __init__(self, path, file_lister=None, ssh_user=UNSET_ACTION_KWD,
+    def __init__(self, source, file_lister=None, ssh_user=UNSET_ACTION_KWD,
                  ssh_host=UNSET_ACTION_KWD, ssh_port=UNSET_ACTION_KWD, ssh_key=UNSET_ACTION_KWD):
-        super(PubkeyAuthenticatedTransferAction, self).__init__(path, file_lister=file_lister)
+        super().__init__(source, file_lister=file_lister)
         self.ssh_user = ssh_user
         self.ssh_host = ssh_host
         self.ssh_port = ssh_port
         self.ssh_key = ssh_key
 
     def to_dict(self):
-        return dict(path=self.path, action_type=self.action_type,
-                    ssh_user=self.ssh_user, ssh_host=self.ssh_host,
-                    ssh_port=self.ssh_port)
+        return self._extend_base_dict(
+            ssh_user=self.ssh_user,
+            ssh_host=self.ssh_host,
+            ssh_port=self.ssh_port
+        )
 
     @contextmanager
     def _serialized_key(self):
@@ -488,7 +558,7 @@ class RsyncTransferAction(PubkeyAuthenticatedTransferAction):
 
     @classmethod
     def from_dict(cls, action_dict):
-        return RsyncTransferAction(path=action_dict["path"],
+        return RsyncTransferAction(source=action_dict["source"],
                                    ssh_user=action_dict["ssh_user"],
                                    ssh_host=action_dict["ssh_host"],
                                    ssh_port=action_dict["ssh_port"],
@@ -510,7 +580,7 @@ class ScpTransferAction(PubkeyAuthenticatedTransferAction):
 
     @classmethod
     def from_dict(cls, action_dict):
-        return ScpTransferAction(path=action_dict["path"],
+        return ScpTransferAction(source=action_dict["source"],
                                  ssh_user=action_dict["ssh_user"],
                                  ssh_host=action_dict["ssh_host"],
                                  ssh_port=action_dict["ssh_port"],
@@ -527,7 +597,7 @@ class ScpTransferAction(PubkeyAuthenticatedTransferAction):
                           self.ssh_port, key_file)
 
 
-class MessageAction(object):
+class MessageAction:
     """ Sort of pseudo action describing "files" store in memory and
     transferred via message (HTTP, Python-call, MQ, etc...)
     """
@@ -560,7 +630,14 @@ class MessageAction(object):
         open(path, "w").write(self.contents)
 
 
-DICTIFIABLE_ACTION_CLASSES = [RemoteCopyAction, RemoteTransferAction, MessageAction, RsyncTransferAction, ScpTransferAction]
+DICTIFIABLE_ACTION_CLASSES = [
+    RemoteCopyAction,
+    RemoteTransferAction,
+    MessageAction,
+    RsyncTransferAction,
+    ScpTransferAction,
+    RemoteObjectStoreCopyAction
+]
 
 
 def from_dict(action_dict):
@@ -572,10 +649,18 @@ def from_dict(action_dict):
     if not target_class:
         message = "Failed to recover action from dictionary - invalid action type specified %s." % action_type
         raise Exception(message)
+    if "source" in action_dict:
+        action_dict.pop("path")  # remove redundant information stored for backward compatibility.
+    elif "path" in action_dict:
+        # legacy message received from older Pulsar client, pop the path from the dict
+        # and convert it to a source.
+        source = {"path": action_dict.pop("path")}
+        action_dict["source"] = source
     return target_class.from_dict(action_dict)
 
 
-class BasePathMapper(object):
+class BasePathMapper:
+    match_type: str
 
     def __init__(self, config):
         action_type = config.get('action', DEFAULT_MAPPED_ACTION)
@@ -600,7 +685,8 @@ class BasePathMapper(object):
 
     def matches(self, path, path_type):
         path_type_matches = path_type in self.path_types
-        return path_type_matches and self._path_matches(path)
+        rval = path_type_matches and self._path_matches(path)
+        return rval
 
     def _extend_base_dict(self, **kwds):
         base_dict = dict(
@@ -613,19 +699,35 @@ class BasePathMapper(object):
         base_dict.update(**kwds)
         return base_dict
 
+    def to_pattern(self):
+        raise NotImplementedError()
+
+
+class PathTypeOnlyMapper(BasePathMapper):
+    match_type = 'path_type_only'
+
+    def __init__(self, config):
+        super().__init__(config)
+
+    def _path_matches(self, path):
+        return True
+
+    def to_dict(self):
+        return self._extend_base_dict()
+
 
 class PrefixPathMapper(BasePathMapper):
     match_type = 'prefix'
 
     def __init__(self, config):
-        super(PrefixPathMapper, self).__init__(config)
+        super().__init__(config)
         self.prefix_path = abspath(config['path'])
 
     def _path_matches(self, path):
-        return path.startswith(self.prefix_path)
+        return path is not None and path.startswith(self.prefix_path)
 
     def to_pattern(self):
-        pattern_str = r"(%s%s[^\s,\"\']+)" % (escape(self.prefix_path), escape(sep))
+        pattern_str = r"({}{}[^\s,\"\']+)".format(escape(self.prefix_path), escape(sep))
         return compile(pattern_str)
 
     def to_dict(self):
@@ -636,11 +738,11 @@ class GlobPathMapper(BasePathMapper):
     match_type = 'glob'
 
     def __init__(self, config):
-        super(GlobPathMapper, self).__init__(config)
+        super().__init__(config)
         self.glob_path = config['path']
 
     def _path_matches(self, path):
-        return fnmatch.fnmatch(path, self.glob_path)
+        return path is not None and fnmatch.fnmatch(path, self.glob_path)
 
     def to_pattern(self):
         return compile(fnmatch.translate(self.glob_path))
@@ -653,12 +755,12 @@ class RegexPathMapper(BasePathMapper):
     match_type = 'regex'
 
     def __init__(self, config):
-        super(RegexPathMapper, self).__init__(config)
+        super().__init__(config)
         self.pattern_raw = config['path']
         self.pattern = compile(self.pattern_raw)
 
     def _path_matches(self, path):
-        return self.pattern.match(path) is not None
+        return path is not None and self.pattern.match(path) is not None
 
     def to_pattern(self):
         return self.pattern
@@ -667,7 +769,7 @@ class RegexPathMapper(BasePathMapper):
         return self._extend_base_dict(path=self.pattern_raw)
 
 
-MAPPER_CLASSES = [PrefixPathMapper, GlobPathMapper, RegexPathMapper]
+MAPPER_CLASSES = [PathTypeOnlyMapper, PrefixPathMapper, GlobPathMapper, RegexPathMapper]
 MAPPER_CLASS_DICT = dict(map(lambda c: (c.match_type, c), MAPPER_CLASSES))
 
 
@@ -676,11 +778,14 @@ def mappers_from_dicts(mapper_def_list):
 
 
 def _mappper_from_dict(mapper_dict):
-    map_type = mapper_dict.get('match_type', DEFAULT_PATH_MAPPER_TYPE)
+    if "path" in mapper_dict:
+        map_type = mapper_dict.get('match_type', DEFAULT_PATH_MAPPER_TYPE)
+    else:
+        map_type = 'path_type_only'
     return MAPPER_CLASS_DICT[map_type](mapper_dict)
 
 
-class FileLister(object):
+class FileLister:
 
     def __init__(self, config):
         self.depth = int(config.get("depth", "0"))
@@ -698,22 +803,23 @@ class FileLister(object):
             while depth > 0:
                 path = dirname(path)
                 depth -= 1
-            return dict([(join(path, f), f) for f in directory_files(path)])
+            return {join(path, f): f for f in directory_files(path)}
 
 
 DEFAULT_FILE_LISTER = FileLister(dict(depth=0))
 
-ACTION_CLASSES = [
+ACTION_CLASSES: List[Type[BaseAction]] = [
     NoneAction,
     RewriteAction,
     TransferAction,
     CopyAction,
     RemoteCopyAction,
     RemoteTransferAction,
+    RemoteObjectStoreCopyAction,
     RsyncTransferAction,
     ScpTransferAction,
 ]
-actions = dict([(clazz.action_type, clazz) for clazz in ACTION_CLASSES])
+actions = {clazz.action_type: clazz for clazz in ACTION_CLASSES}
 
 
 __all__ = (
