@@ -1,35 +1,37 @@
 import logging
 import os
 
-from six import string_types
-
 from pulsar.managers.util.pykube_util import (
+    delete_job,
     ensure_pykube,
     find_job_object_by_name,
     find_pod_object_by_name,
     galaxy_instance_id,
     Job,
     job_object_dict,
-    produce_unique_k8s_job_name,
+    produce_k8s_job_prefix,
     pull_policy,
     pykube_client_from_dict,
-    stop_job,
 )
 from .action_mapper import (
     actions,
     path_type,
 )
 from .amqp_exchange import ACK_FORCE_NOACK_KEY
-from .decorators import parseJson
-from .decorators import retry
+from .decorators import (
+    parseJson,
+    retry,
+)
 from .destination import submit_params
 from .job_directory import RemoteJobDirectory
 from .setup_handler import build as build_setup_handler
-from .util import copy
-from .util import ensure_directory
-from .util import json_dumps
-from .util import json_loads
-from .util import to_base64_json
+from .util import (
+    copy,
+    ensure_directory,
+    json_dumps,
+    json_loads,
+    to_base64_json,
+)
 
 log = logging.getLogger(__name__)
 
@@ -53,7 +55,7 @@ class OutputNotFoundException(Exception):
         return "No remote output found for path %s" % self.path
 
 
-class BaseJobClient(object):
+class BaseJobClient:
 
     def __init__(self, destination_params, job_id):
         destination_params = destination_params or {}
@@ -127,10 +129,10 @@ class JobClient(BaseJobClient):
     """
 
     def __init__(self, destination_params, job_id, job_manager_interface):
-        super(JobClient, self).__init__(destination_params, job_id)
+        super().__init__(destination_params, job_id)
         self.job_manager_interface = job_manager_interface
 
-    def launch(self, command_line, dependencies_description=None, env=[], remote_staging=[], job_config=None):
+    def launch(self, command_line, dependencies_description=None, env=None, remote_staging=None, job_config=None, dynamic_file_sources=None):
         """
         Queue up the execution of the supplied `command_line` on the remote
         server. Called launch for historical reasons, should be renamed to
@@ -161,6 +163,8 @@ class JobClient(BaseJobClient):
             # before queueing.
             setup_params = _setup_params_from_job_config(job_config)
             launch_params['setup_params'] = json_dumps(setup_params)
+        if dynamic_file_sources is not None:
+            launch_params["dynamic_file_sources"] = json_dumps(dynamic_file_sources)
         return self._raw_execute("submit", launch_params)
 
     def full_status(self):
@@ -213,7 +217,7 @@ class JobClient(BaseJobClient):
         # action type == 'message' should either copy or transfer
         # depending on default not just fallback to transfer.
         if action_type in ['transfer', 'message']:
-            if isinstance(contents, string_types):
+            if isinstance(contents, str):
                 contents = contents.encode("utf-8")
             message = "Uploading path [%s] (action_type: [%s])"
             log.debug(message, path, action_type)
@@ -250,7 +254,9 @@ class JobClient(BaseJobClient):
         else:
             raise Exception("Unknown output_type %s" % output_type)
 
-    def _raw_execute(self, command, args={}, data=None, input_path=None, output_path=None):
+    def _raw_execute(self, command, args=None, data=None, input_path=None, output_path=None):
+        if args is None:
+            args = {}
         return self.job_manager_interface.execute(command, args, data, input_path, output_path)
 
     def _fetch_output(self, path, name=None, check_exists_remotely=False, action_type='transfer'):
@@ -296,7 +302,7 @@ class JobClient(BaseJobClient):
 class BaseMessageJobClient(BaseJobClient):
 
     def __init__(self, destination_params, job_id, client_manager):
-        super(BaseMessageJobClient, self).__init__(destination_params, job_id)
+        super().__init__(destination_params, job_id)
         if not self.job_directory:
             error_message = "Message-queue based Pulsar client requires destination define a remote job_directory to stage files into."
             raise Exception(error_message)
@@ -311,7 +317,7 @@ class BaseMessageJobClient(BaseJobClient):
             raise Exception("full_status() called before a final status was properly cached with cilent manager.")
         return full_status
 
-    def _build_setup_message(self, command_line, dependencies_description, env, remote_staging, job_config):
+    def _build_setup_message(self, command_line, dependencies_description, env, remote_staging, job_config, dynamic_file_sources):
         """
         """
         launch_params = dict(command_line=command_line, job_id=self.job_id)
@@ -325,6 +331,7 @@ class BaseMessageJobClient(BaseJobClient):
         if remote_staging:
             launch_params['remote_staging'] = remote_staging
             launch_params['remote_staging']['ssh_key'] = self.ssh_key
+        launch_params['dynamic_file_sources'] = dynamic_file_sources
         if job_config and self.setup_handler.local:
             # Setup not yet called, job properties were inferred from
             # destination arguments. Hence, must have Pulsar setup job
@@ -345,7 +352,7 @@ class BaseMessageJobClient(BaseJobClient):
 
 class MessageJobClient(BaseMessageJobClient):
 
-    def launch(self, command_line, dependencies_description=None, env=[], remote_staging=[], job_config=None):
+    def launch(self, command_line, dependencies_description=None, env=None, remote_staging=None, job_config=None, dynamic_file_sources=None):
         """
         """
         launch_params = self._build_setup_message(
@@ -354,6 +361,7 @@ class MessageJobClient(BaseMessageJobClient):
             env=env,
             remote_staging=remote_staging,
             job_config=job_config,
+            dynamic_file_sources=dynamic_file_sources,
         )
         response = self.client_manager.exchange.publish("setup", launch_params)
         log.info("Job published to setup message queue: %s", self.job_id)
@@ -372,11 +380,11 @@ class MessageJobClient(BaseMessageJobClient):
 class MessageCLIJobClient(BaseMessageJobClient):
 
     def __init__(self, destination_params, job_id, client_manager, shell):
-        super(MessageCLIJobClient, self).__init__(destination_params, job_id, client_manager)
+        super().__init__(destination_params, job_id, client_manager)
         self.remote_pulsar_path = destination_params["remote_pulsar_path"]
         self.shell = shell
 
-    def launch(self, command_line, dependencies_description=None, env=[], remote_staging=[], job_config=None):
+    def launch(self, command_line, dependencies_description=None, env=None, remote_staging=None, job_config=None, dynamic_file_sources=None):
         """
         """
         launch_params = self._build_setup_message(
@@ -385,11 +393,12 @@ class MessageCLIJobClient(BaseMessageJobClient):
             env=env,
             remote_staging=remote_staging,
             job_config=job_config,
+            dynamic_file_sources=dynamic_file_sources,
         )
         base64_message = to_base64_json(launch_params)
         submit_command = os.path.join(self.remote_pulsar_path, "scripts", "submit.bash")
         # TODO: Allow configuration of manager, app, and ini path...
-        self.shell.execute("nohup %s --base64 %s &" % (submit_command, base64_message))
+        self.shell.execute("nohup {} --base64 {} &".format(submit_command, base64_message))
 
     def kill(self):
         # TODO
@@ -400,19 +409,19 @@ class MessageCoexecutionPodJobClient(BaseMessageJobClient):
 
     def __init__(self, destination_params, job_id, client_manager):
         ensure_pykube()
-        super(MessageCoexecutionPodJobClient, self).__init__(destination_params, job_id, client_manager)
+        super().__init__(destination_params, job_id, client_manager)
         self.instance_id = galaxy_instance_id(destination_params)
         self.pulsar_container_image = destination_params.get("pulsar_container_image", "galaxy/pulsar-pod-staging:0.13.0")
         self._default_pull_policy = pull_policy(destination_params)
 
-    # FIXME: pulsar/client/client.py:404:5: C901 'MessageCoexecutionPodJobClient.launch' is too complex (13)
-    def launch(  # noqa: C901
+    def launch(
         self,
         command_line,
         dependencies_description=None,
-        env=[],
-        remote_staging=[],
+        env=None,
+        remote_staging=None,
         job_config=None,
+        dynamic_file_sources=None,
         container_info=None,
         pulsar_app_config=None
     ):
@@ -424,6 +433,7 @@ class MessageCoexecutionPodJobClient(BaseMessageJobClient):
             env=env,
             remote_staging=remote_staging,
             job_config=job_config,
+            dynamic_file_sources=dynamic_file_sources,
         )
         container = None
         guest_ports = None
@@ -469,7 +479,7 @@ class MessageCoexecutionPodJobClient(BaseMessageJobClient):
         base64_message = to_base64_json(launch_params)
         base64_app_conf = to_base64_json(pulsar_app_config)
 
-        job_name = self._k8s_job_name
+        k8s_job_prefix = self._k8s_job_prefix
         params = self.destination_params
 
         pulsar_container_image = self.pulsar_container_image
@@ -517,7 +527,7 @@ class MessageCoexecutionPodJobClient(BaseMessageJobClient):
 
         template = {
             "metadata": {
-                "labels": {"app": job_name},
+                "labels": {"app": k8s_job_prefix},
             },
             "spec": {
                 "volumes": volumes,
@@ -526,25 +536,24 @@ class MessageCoexecutionPodJobClient(BaseMessageJobClient):
             }
         }
         spec = {"template": template}
-        if "k8s_walltime_limit" in params:
-            spec["activeDeadlineSeconds"] = int(params["k8s_walltime_limit"])
-        k8s_job_obj = job_object_dict(params, job_name, spec)
+        spec.update(self._job_spec_params(params))
+        k8s_job_obj = job_object_dict(params, k8s_job_prefix, spec)
         pykube_client = self._pykube_client
         job = Job(pykube_client, k8s_job_obj)
         job.create()
 
     def kill(self):
-        job_name = self._k8s_job_name
+        job_name = self._k8s_job_prefix
         pykube_client = self._pykube_client
         job = find_job_object_by_name(pykube_client, job_name)
         if job:
             log.info("Kill k8s job with name %s" % job_name)
-            stop_job(job)
+            delete_job(job)
         else:
             log.info("Attempted to kill k8s job but it is unavailable.")
 
     def job_ip(self):
-        job_name = self._k8s_job_name
+        job_name = self._k8s_job_prefix
         pykube_client = self._pykube_client
         pod = find_pod_object_by_name(pykube_client, job_name)
         if pod:
@@ -563,10 +572,18 @@ class MessageCoexecutionPodJobClient(BaseMessageJobClient):
         return pykube_client_from_dict(self.destination_params)
 
     @property
-    def _k8s_job_name(self):
+    def _k8s_job_prefix(self):
         job_id = self.job_id
-        job_name = produce_unique_k8s_job_name(app_prefix="pulsar", job_id=job_id, instance_id=self.instance_id)
-        return job_name
+        job_prefix = produce_k8s_job_prefix(app_prefix="pulsar", job_id=job_id, instance_id=self.instance_id)
+        return job_prefix
+
+    def _job_spec_params(self, params):
+        spec = {}
+        if "k8s_walltime_limit" in params:
+            spec["activeDeadlineSeconds"] = int(params["k8s_walltime_limit"])
+        if "k8s_job_ttl_secs_after_finished" in params and params.get("k8s_cleanup_job") != "never":
+            spec["ttlSecondsAfterFinished"] = int(params["k8s_job_ttl_secs_after_finished"])
+        return spec
 
     def _pulsar_container_resources(self, params):
         return self._container_resources(params, container='pulsar')
@@ -595,7 +612,7 @@ class InputCachingJobClient(JobClient):
     """
 
     def __init__(self, destination_params, job_id, job_manager_interface, client_cacher):
-        super(InputCachingJobClient, self).__init__(destination_params, job_id, job_manager_interface)
+        super().__init__(destination_params, job_id, job_manager_interface)
         self.client_cacher = client_cacher
 
     @parseJson()
@@ -643,6 +660,8 @@ def _setup_params_from_job_config(job_config):
     tool_id = job_config.get("tool_id", None)
     tool_version = job_config.get("tool_version", None)
     preserve_galaxy_python_environment = job_config.get("preserve_galaxy_python_environment", None)
+    # use_metadata ignored post Pulsar 0.14.12+ but keep setting it for older Pulsar's that
+    # had hacks for pre-2017 Galaxies.
     return dict(
         job_id=job_id,
         tool_id=tool_id,
